/*
 * Copyright (C) 2014-2015 Darron Broad
 * All rights reserved.
 *
 * This file is part of Mork Microchip PIC ICSP.
 *
 * Mork Microchip PIC ICSP is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * Mork Microchip PIC ICSP is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Mork Microchip PIC ICSP. If not, see http://www.gnu.org/licenses/
 */

#include "mork.h"

/******************************************************************************
  PANEL WRITING
 *****************************************************************************/

/*
 * WRITE PANEL
 *
 *  mode:
 *  	PIC_PANEL_BEGIN
 *  		data1: 	memory region
 *  		data2:	memory size
 *
 *  	PIC_PANEL_UPDATE
 *  		data1:	memory address
 *  		data2:	memory data
 *
 *  	PIC_PANEL_END
 *  		data1:	void
 *  		data2:	void
 */
int16_t
pic_write_panel(uint8_t mode, uint32_t data1, uint32_t data2)
{
	static uint32_t panel_region;
	static uint32_t panel_address;
	static uint32_t panel_size;
	static uint32_t write_pending;
	static uint8_t panel[512];

	if (mode == PIC_PANEL_BEGIN) {
		panel_region = (uint8_t)data1;		/* MEMORY REGION */
		panel_size = data2;			/* MEMORY SIZE 128 or 512 bytes for PIC32 */
		if (panel_size > 512) {
			return -1;
		}
		write_pending = 0;
		memset(panel, 0xFF, panel_size);
	}
	else if (mode == PIC_PANEL_UPDATE) {
		uint32_t address, new_address, boundary, mask;

		boundary = 0 - panel_size;
		mask = panel_size - 1;
		address = data1;			/* MEMORY ADDRESS */
		new_address = address & boundary;
		if (new_address != panel_address) {
			if (write_pending) {
				pic32_write_panel(panel_region, panel_address, panel, panel_size);
				write_pending = 0;
				memset(panel, 0xFF, panel_size);
			}
			panel_address = new_address;
		}
		panel[address & mask] = (uint8_t)data2;	/* MEMORY DATA */
		write_pending++;
	}
	else if (mode == PIC_PANEL_END) {
		if (write_pending) {
			pic32_write_panel(panel_region, panel_address, panel, panel_size);
			write_pending = 0;
		}
	}
	return 0;
}
