/*
 * Copyright (C) 2014-2015 Darron Broad
 * All rights reserved.
 *
 * This file is part of Mork Microchip PIC ICSP.
 *
 * Mork Microchip PIC ICSP is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * Mork Microchip PIC ICSP is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Mork Microchip PIC ICSP. If not, see http://www.gnu.org/licenses/
 */

#include "mork.h"

/*******************************************************************************
 *
 * STK500V2 emulator
 *
 ******************************************************************************/

/* I/O buffer */
uint8_t stk_msg[STK_LEN + /* Checksum */ 1];

/* Packet sequence number */
uint8_t stk_seqnum;

/* I/O data row address */
uint32_t stk_addr;

/* Sign-on signature */
const uint8_t stk_signon_msg[11] = {
	STK_CMD_SIGN_ON, STK_STATUS_CMD_OK,
	/* SIG LENGTH   */ 0x08,
	/* SIG STK500V2 */ 0x53,0x54,0x4B,0x35,0x30,0x30,0x5F,0x32
};

/* Programming mode NONE/PROGRAM/VERIFY */
static int stk_progmode;

/*******************************************************************************
 *
 * AVR068: STK500V2 Communication Protocol
 *
 *  Receive Packet
 *
 * Message Format & Protocol Layer State Table
 *
 * 2591C-AVR-06/06 Pages 2 & 3
 *
 ******************************************************************************/
int16_t
stk500v2_recv(void)
{
	int rc;
	uint8_t sum = 0;
	uint16_t msgsize, i;

	/* Get message start */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}

	/* Validate message start */
	if (rc != STK_MESSAGE_START) {
		return STK_PROTOCOL_ERROR;
	}

	/* Message start checksum */
	sum ^= rc;

	/* Get sequence number */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}

	/* Echo sequence number */
	stk_seqnum = rc;

	/* Sequence number checksum */
	sum ^= rc;

	/* Get message size 1 */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}
	msgsize = rc << 8;

	/* Message size 1 checksum */
	sum ^= rc;

	/* Get message size 2 */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}
	msgsize |= rc;

	/* Message size 2 checksum */
	sum ^= rc;

	/* Validate message size */
	if (msgsize > STK_LEN) {
		return STK_PROTOCOL_ERROR;
	}

	/* Get token */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}

	/* Validate token */
	if (rc != STK_TOKEN) {
		return STK_PROTOCOL_ERROR;
	}

	/* Token checksum */
	sum ^= rc;

	/* Get message data */
	for (i = 0; i < msgsize; ++i) {
		rc = uart_getchar();
		if (rc < 0) {
			return rc;
		}
		stk_msg[i] = rc;

		/* Message checksum */
		sum ^= rc;
	}

	/* Get checksum */
	rc = uart_getchar();
	if (rc < 0) {
		return rc;
	}

	/* Validate packet checksum */
	if (sum != rc) {
		return STK_PROTOCOL_ERROR;
	}

	/* Return with number of bytes in message */
	return msgsize;
}

/*******************************************************************************
 *
 * AVR068: STK500V2 Communication Protocol
 *
 *  Send Packet
 *
 * Message Format & General Commands
 *
 * 2591C-AVR-06/06 Pages 2 & 4
 *
 ******************************************************************************/
int16_t
stk500v2_send(uint16_t msgsize)
{
	int rc;
	uint8_t header[5], sum = 0;
	uint16_t i;

	/* Send header */
	header[0] = STK_MESSAGE_START;
	header[1] = stk_seqnum;
	header[2] = msgsize >> 8;
	header[3] = msgsize & 0xFF;
	header[4] = STK_TOKEN;
	for (i = 0; i < 5; ++i) {
		rc = uart_putchar(header[i]);
		if (rc < 0) {
			return rc;
		}
		sum ^= header[i];
	}

	/* Message checksum */
	for (i = 0; i < msgsize; ++i)
		sum ^= stk_msg[i];

	/* Packet checksum */
	stk_msg[msgsize] = sum;

	/* Send message & Checksum */
	for (i = 0; i <= msgsize; ++i) {
		rc = uart_putchar(stk_msg[i]);
		if (rc < 0) {
			return rc;
		}
	}

	/* Flush output */
	uart_flush();

	/* Return with number of bytes in message */
	return msgsize;
}

/******************************************************************************
 *
 * STK500V2 listener
 *
 *****************************************************************************/

/******************************************************************************
 * CLEAN UP DEAD SESSION
 */
static void
stk500v2_cleanup(void)
{
	if (stk_progmode == STK_PROGMODE_WRITE)
		pic32_program_end(1);
	else if (stk_progmode == STK_PROGMODE_READ)
		pic32_standby();

	stk_progmode = STK_PROGMODE_NONE;
}

/******************************************************************************
 * RESPOND WITH COMMAND OKAY
 */
static inline void
stk500v2_ok(void)
{
	stk_msg[1] = STK_STATUS_CMD_OK;
	stk500v2_send(2);
}

/******************************************************************************
 * RESPOND WITH COMMAND FAILED
 */
static inline void
stk500v2_failed(void)
{
	stk_msg[1] = STK_STATUS_CMD_FAILED;
	stk500v2_send(2);
}

/******************************************************************************
 * RESPOND TO COMMAND SIGN-ON
 */
static inline void
stk500v2_signon(void)
{
	memcpy((void *)stk_msg, (void *)stk_signon_msg, sizeof(stk_signon_msg));
	stk500v2_send(sizeof(stk_signon_msg));
}

/******************************************************************************
 * RESPOND TO COMMAND SET PARAMETER
 */
static inline void
stk500v2_set_parameter(void)
{
	stk500v2_ok();
}

/******************************************************************************
 * RESPOND TO COMMAND GET PARAMETER
 */
static inline void
stk500v2_get_parameter(void)
{
	stk_msg[1] = STK_STATUS_CMD_OK;
	stk_msg[2] = /* Value */ 0;
	stk500v2_send(3);
}

/******************************************************************************
 * RESPOND TO COMMAND SPI MULTI
 */
static inline void
stk500v2_spi_multi(void)
{
	if (stk_msg[1] == 4 && stk_msg[2] == 4 && stk_msg[3] == 0) {
		stk_msg[1] = STK_STATUS_CMD_OK;
		stk_msg[2] = 0;
		stk_msg[3] = 0;
		if (stk_msg[4] == 0x30) {
			/* Signature query */
			stk_msg[4] = 0;
			switch (stk_msg[6]) {
			case 0: stk_msg[5] = 0x50; /* chipKIT PIC32 [0] */
				break;
			case 1: stk_msg[5] = 0x49; /* chipKIT PIC32 [1] */
				break;
			case 2: stk_msg[5] = 0x43; /* chipKIT PIC32 [2] */
				break;
			}
		} else {
			stk_msg[4] = 0;
			stk_msg[5] = 0;
		}
		stk_msg[6] = STK_STATUS_CMD_OK;
		stk500v2_send(7);
	} else {
		stk500v2_failed();
	}
}

/******************************************************************************
 * RESPOND TO COMMAND CHIP ERASE ISP
 */
static inline void
stk500v2_chip_erase_isp(void)
{
	stk500v2_cleanup();

	if (pic32_read_config_memory() == 0) {
		pic32_bulk_erase();
		stk500v2_ok();
		return;
	}
	stk500v2_failed();
}

/******************************************************************************
 * RESPOND TO COMMAND ENTER PROGMODE ISP
 */
static inline void
stk500v2_enter_progmode_isp(void)
{
	stk500v2_cleanup();

	if (pic32_read_config_memory() == 0) {
		stk500v2_ok();
		return;
	}
	stk500v2_failed();
}

/******************************************************************************
 * RESPOND TO COMMAND LOAD ADDRESS
 */
static inline void
stk500v2_load_address(void)
{
	stk_addr = (uint32_t)stk_msg[1] << 24 | (uint32_t)stk_msg[2] << 16 |
		(uint32_t)stk_msg[3] << 8 | (uint32_t)stk_msg[4];
	if (stk_addr & 0x80000000) {
		stk_addr <<= 1;                 /* 32-BIT EXTENDED ADDRESS */
	} else {
		stk_addr <<= 1;                 /* RELATIVE ADDRESS */
		stk_addr += PIC32_CODE;
	}
	stk500v2_ok();
}

/******************************************************************************
 * RESPOND TO COMMAND PROGRAM FLASH ISP
 */
static inline void
stk500v2_program_flash_isp(void)
{
	static uint32_t region;
	pic_data pdata;

	if (stk_progmode == STK_PROGMODE_READ) {
		pic32_standby();
	}
	if (stk_progmode != STK_PROGMODE_WRITE) {
		if (pic32_read_config_memory() == 0) {
			pic32_program_begin();
			stk_progmode = STK_PROGMODE_WRITE;
			region = PIC_REGIONNOTSUP;
		} else  {
			stk_progmode = STK_PROGMODE_NONE;
			stk500v2_failed();
			return;
		}
	}

	pdata.nbytes = (uint16_t)stk_msg[1] << 8 | (uint16_t)stk_msg[2];

	if (pdata.nbytes <= PIC_BYTLEN) {
		pdata.address = stk_addr;
		pdata.bytes = &stk_msg[10];
		stk_addr += pdata.nbytes;
		region = pic32_program_data(region, &pdata);
		stk500v2_ok();
	} else  {
		stk500v2_failed();
	}
}

/******************************************************************************
 * RESPOND TO COMMAND READ FLASH ISP
 */
static inline void
stk500v2_read_flash_isp(void)
{
	static uint32_t region;
	uint32_t fail;
	pic_data pdata;

	if (stk_progmode == STK_PROGMODE_WRITE) {
		pic32_program_end(1);
	}
	if (stk_progmode != STK_PROGMODE_READ) {
		if (pic32_read_config_memory() == 0) {
			pic32_program_begin();
			stk_progmode = STK_PROGMODE_READ;
			region = PIC_REGIONNOTSUP;
		} else  {
			stk_progmode = STK_PROGMODE_NONE;
			stk500v2_failed();
			return;
		}
	}

	pdata.nbytes = (uint16_t)stk_msg[1] << 8 | (uint16_t)stk_msg[2];

	if (pdata.nbytes <= PIC_BYTLEN) {
		pdata.address = stk_addr;
		pdata.bytes = &stk_msg[2];
		stk_addr += pdata.nbytes;
		region = pic32_verify_data(region, &pdata, &fail);
		stk_msg[1] = STK_STATUS_CMD_OK;
		stk_msg[2 + pdata.nbytes] = STK_STATUS_CMD_OK;
		stk500v2_send(3 + pdata.nbytes);
	} else {
		stk500v2_failed();
	}
}

/******************************************************************************
 * RESPOND TO COMMAND LEAVE PROGMODE ISP
 */
static inline void
stk500v2_leave_progmode_isp(void)
{
	stk500v2_cleanup();
	stk500v2_ok();
}

/******************************************************************************
 * LISTEN FOR REMOTE PROGRAMMING
 */
int16_t
stk500v2_listen(void)
{
	int16_t rc;

	/* Init */
	stk_addr = 0;
	stk_progmode = STK_PROGMODE_NONE;

	while (1) {
		rc = stk500v2_recv();
		if (rc == STK_PROTOCOL_ERROR) {
			continue;
		}
		if (rc == /* Empty message */ 0) {
			continue;
		}
		if (rc < 0) {
			stk500v2_cleanup();
			break;
		}
		if (stk_msg[0] == STK_CMD_LEAVE_PROGMODE_ISP) {
			stk500v2_leave_progmode_isp();
			break;
		}
		switch (stk_msg[0]) {
		default:stk500v2_failed();
			break;
		case STK_CMD_SIGN_ON:
			stk500v2_signon();
			break;
		case STK_CMD_SET_PARAMETER:
			stk500v2_set_parameter();
			break;
		case STK_CMD_GET_PARAMETER:
			stk500v2_get_parameter();
			break;
		case STK_CMD_SPI_MULTI:
			stk500v2_spi_multi();
			break;
		case STK_CMD_CHIP_ERASE_ISP:
			stk500v2_chip_erase_isp();
			break;
		case STK_CMD_ENTER_PROGMODE_ISP:
			stk500v2_enter_progmode_isp();
			break;
		case STK_CMD_LOAD_ADDRESS:
			stk500v2_load_address();
			break;
		case STK_CMD_PROGRAM_FLASH_ISP:
			stk500v2_program_flash_isp();
			break;
		case STK_CMD_READ_FLASH_ISP:
			stk500v2_read_flash_isp();
			break;
		}
	}

	return rc;
}
